/****************************************************************************
 *                                                                          *
 * This is the YACC file that defines a BNF grammar for the Scene           *
 * Description Language (SDL).						    *
 * 									    *
 * Some notational conventions that are used:				    *
 * 	Z1item - Zero or 1 occurences of item.				    *
 *	ZMitem - Zero or More occurences of item.			    *
 *	OMitem - One or More occurences of item.			    *
 *	work_*lst - a Pointer to a list; where "*" can be "m" (matrix)	    *
 *		"v3" (3-vectors) etc					    *
 *                                                                          *
 *                                                                          *
 * ALSO, don't forget:							    *
 *	When parameter counting in the non-terminals (trying to work out    *
 *	exactly what "$7", or whatever, is supposed to be - any little	    *
 *	bits of C code that appear before $7 within the definition will up  *
 *	the parameter count by 1.					    *
 *									    *
 *	eg:	SPHERE					$1		    *
 *		     Z1name             		$2		    *
 *		     { ...some C code... }		$3		    *
 *		     NUMBER				$4		    *
 *		     .... etc ....			...		    *
 *                                                                          *
 *				Neil Gatenby, 14-07-92			    *
 * Properties:								    *
 *	Sets of surface properties which are given a name by the user will  *
 *	be added to a list of defined surface properties, which can be	    *
 *	instantiated later in the file - those which are not named are	    *
 *	given a default name (to aid with debugging etc.) but are NOT added *
 *	to the list.							    *
 ****************************************************************************/


%{
#include <stdio.h>
#include <sys/types.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>
#include "msdl_vec.h"
#include "msdl_typ.h"
#include "msdl_pro.h"

/*=== S D L     V A R I A B L E S ====================================*/
int		lineno = 1;
char		work_char;
Tboolean	work_defining = FALSE;
ptr2mlist	work_DMlst = NULL;	/* list of Defined Matrices */
ptr2Rlist	work_Rlst = NULL;	/* list of reals (Tfloats) */
ptr2v3list	work_v3lst = NULL;	/* list of Tpt3s */
ptr2v4list	work_v4lst = NULL;	/* list of Tpt4s */
Tsprops		work_sprops;
ptr2sprops	work_SPlst = NULL;
ptr2llist	work_llst = NULL;
Tstring		work_string;
Tmatrix		work_TM1,		/* working matrix */
		work_TM2;		/* 2nd working matrix */
ptr2obj		work_Pobj  = NULL,	/* current parent object */
		work_DOlst = NULL,	/* Defined Objects list */
		work_IOlst = NULL;	/* Instan' Objects list - the scene */
Tptchs		work_patches;
TThetaPhi	work_theta,
		work_phi;
ptr2face	work_face = NULL;
ptr2flist	work_flst = NULL;
Tview		*work_view= NULL;		/* ptr 2 the view */
decMLS ( ptr2obj, Pstack );
decMLS ( Tsprops, SPstack );
/* if U change this decSTR, change the EXTERN decs in objects.c as well */
decSTR ( Sstack );	
decTMS ( Mstack );

/*=== U S E R     V A R I A B L E S =========================================*/


%}

%union {
    int		cmd;
    Tstring	str;				/* strings */
    Tfloat	num_R;				/* real numbers */
    int		num_Z;				/* integers */
    Tpt3	pt3;
    Tpt4	pt4;
    Tboolean	bool;
    TThetaPhi	thph;
    ptr2obj	Optr;
}

%token	<cmd>	
	VIEW
	FROM
	TOWARDS
	VPN
	UP
	WINDOW
	FRONT
	BACK
	PARALLEL
	OPENBRACE
	CLOSEBRACE
	OPENSQ
	CLOSESQ
	POINT
	DIRECTIONAL
	SPOT
	DEFPROPS
	PROPS
	INSTPROPS
	NAME
	EMISSION
	ARC
	DRC
	SRC
	TRC
	PHEXP
	INDEX
	TEXTURE
	TEXFILE
	DEFOBJ
	COMPOBJ
	INSTOBJ
	POLYGON
	TRIANGLE
	VNORM
	FNORM
	SPHERE
	TPZ
	THETA
	PHI
	CYLINDER
	BOTTOM
	TOP
	CAPPED
	TZ
	CONE
	DISC
	BOX
	IN
	NURBS
	ORDER
	KNOTS
	CPTS
	POLYHEDRON
	DEFTRANS
	TRANS
	INSTTRANS
	SHIFT
	ROTATE
	SCALE
	MATRIX
	PATCHES
	WALK
	WALK2
	END
%token	<num_R> 	NUMBER
%token	<str>		ID
%type	<str>		Z1name
%type	<pt3>		point3
			Z1up
			Z1tpz
%type	<pt4>		point4
%type	<num_Z>		OMpoint3
			OMpoint4
			OMreals
			OMvlists
			Z1vnorm
			Z1fnorm
%type	<bool>		Z1file
			Z1capped
			Z1in
%type	<thph>		Z1theta
			Z1phi
%type	<Optr>		object
			def_obj_blk
			inst_obj_blk
			obj_blk
			obj_prims
			OMiob_op
			iob_op
			OMiob_ob
			iob_ob
			

%%

/*===========================================================================*
 * Non-terminals in the grammar are labelled (using comments) as lying in a  *
 * particular 'level' - this refers to 'how far into' the grammar they are:  *
 * For example, "scene" is at level 0 - all the nonterminals making up scene *
 * are at level 1; all the nonterminals making up level N are at level N+1.  *
 * 									     *
 * Within the levels, the nonterminals are ordered according to the order in *
 * which they were mentioned at the previous level.			     *
 *---------------------------------------------------------------------------*/


/*===========================  L E V E L    0  ==============================*/

scene:		  view_blk
		  scene_blk
		;

/*===========================  L E V E L    1  ==============================*/

view_blk:	  /* nothing */
		| VIEW
			FROM
			point3		/* $3 */
			{ work_view=NEW( Tview );	/* it needs some space !!*/
			  work_view->from = $3; }
			TOWARDS
			point3		/* $6 */
			{ work_view->towards = $6; }
			Z1vpn
			UP
			point3		/* $10 */
			WINDOW
			point4		/* $12 */
			Z1front
			Z1back
			Z1parallel
			END
			{
			  /* code to say is has been defined */
			  work_view->up = $10;
			  work_view->window.tr.u = $12.x;
			  work_view->window.bl.u = $12.y;
			  work_view->window.tr.v = $12.z;
			  work_view->window.bl.v = $12.w;
			}
		  ;

/* As many light, property, trans &/or object definitions as I want */
scene_blk:		  /* nothing */
		| scene_blk l_dp_dt_o
		  { cleanup();	/* free everything in sight */ }
		  ;

/*===========================  L E V E L    2  ==============================*/

/* Light, or Def_Props_blk, or Def_Trans_blk, or Object */
l_dp_dt_o:	  light
		| def_props_blk
		| def_trans_blk
		| object 
		| debug
		  ;

/*===========================  L E V E L    3  ==============================*/

light:		  POINT
			point3			/* The light's location */
			point3			/* The light's intensity */
		{ if ( $3.x < 0.0 || $3.y < 0.0 || $3.z < 0.0 )
		      yyerror ( "-ve intensity in POINT light" );
		  add_pt_light ( &($2), (Tspectra *)(&($3)) );
		}
		| DIRECTIONAL
			point3			/* Light is along this dir */
			point3			/* The light's intensity */
		{ if ( $3.x < 0.0 || $3.y < 0.0 || $3.z < 0.0 )
		      yyerror ( "-ve intensity in DIRECTIONAL light" );
		  add_dir_light ( &($2), (Tspectra *)(&($3)) );
		}
		| SPOT
			point3			/* The light's location */
			point3			/* pointing towards */
			point3			/* intensity */
			NUMBER			/* exponent for cos term */
			NUMBER			/* cutoff - in degrees */
		{ if ( $4.x < 0.0 || $4.y < 0.0 || $4.z < 0.0 )
		      yyerror ( "-ve intensity in SPOT light" );
		  if ( OOR($6,0.0,90.0) )	/* OOR; macro; Out Of Range */
		      yyerror ( "Cutoff out of range in SPOT light" );
		  add_spt_light ( &($2), &($3), (Tspectra *)(&($4)),
				 (int)$5, (Tfloat)$6 );
		}
		  ;

def_props_blk:	  DEFPROPS
			ID
			{
			  strcpy ( work_string, $2 );
			  get_props ( &work_sprops, NULL );
			}
			ZMipb
			ZMpp
			END
			{ add_props( &work_sprops, work_string ); }
		  ;

def_trans_blk:	  DEFTRANS
			ID
			{
			  strcpy ( work_string, $2 );
			  unit_matrix ( work_TM1 );
			}
			ZMitb
			ZMtp
			END
			{
			  popSTR ( Sstack, work_string );
			  add_trans ( work_TM1, work_string );
			}
			;			

object:		  def_obj_blk
		| obj_blk
		| inst_obj_blk
		  ;

/*===========================  L E V E L    4  ==============================*/

 point3:		NUMBER NUMBER NUMBER
			{ $$.x = $1,
			  $$.y = $2,
			  $$.z = $3;
			}
		  ;
						   
/* Zero or More Instantiate Properties Blocks (ZMIPB) */
ZMipb:		  /* nothing */
		| ZMipb inst_props_blk
		  ;

/* Zero or More Property Primitives (ZMPP) */
ZMpp:		  /* nothing */
		| ZMpp props_prims
		  ;

/* Zero or More Inst_Trans_Blks (ZMITB) */
ZMitb:		  /* nothing */
		| ZMitb inst_trans_blk
		  ;

/* Zero or More Trans_Prims (ZMTP) */
ZMtp:		  /* nothing */
		| ZMtp trans_prims
		  ;

/* Zero or One viewing plane normals */
Z1vpn:		  /* nothing */
		  { Tpt3 tmp;

		    SubVec ( work_view->from, work_view->towards, tmp );
		    normalize ( &tmp );
		    work_view->vpn = tmp;
		  }
		| VPN point3
		  { Tpt3 tmp = $2;

		    normalize ( &tmp );
		    work_view->vpn = tmp;
		  }
		;   

/* zero or one front distances */
Z1front:	  /* nothing */
		  { work_view->fpd = 10000; }
		| FRONT NUMBER
		  { work_view->fpd = $2; }
		;

/* zero or one back distances */
Z1back:		  /* nothing */
		  { work_view->bpd = 0; }
		| BACK NUMBER
		  { work_view->bpd = $2; }
		;

/* zero or one specifying whether perspective */
Z1parallel:	  /* nothing */
		  { work_view->persp = TRUE; }
		| PARALLEL
		  { work_view->persp = FALSE; }
		;


/*-----
if [CASE 1] there's > ONE obj_prim in object_centre, then treat ID as
a parent to everything in OMiob_op,
o/w [CASE 2] just assign the name and bung it on the DO list
 ------*/

def_obj_blk:	  DEFOBJ
			ID
			{
			  pushSTR ( Sstack, $2 );
			  work_defining = TRUE;
			  unit_matrix ( work_TM1 );
			}
			options
			{
			  pushTMS ( Mstack, work_TM1 );
			  pushMLS ( SPstack, work_sprops );
			}
			OMiob_op
			END
			{ 
			  if ( $6->nsib != $6 ) {
			      /*--- [CASE 1] OMiob_op is > 1 objects */
			      ptr2obj	loop;

			      work_Pobj = NEW(Tobject);
			      work_Pobj->parts = $6;
			      popMLS ( SPstack, work_Pobj->sprops );
			      popSTR ( Sstack, work_Pobj->name );

			      /* sort out trans matrix */
			      popTMS ( Mstack, work_Pobj->trans );

			      /* point all parts @ work_Pobj */
			      loop = $6;
			      do {
				  loop->partof = work_Pobj;
				  loop = loop->nsib;
			      } while ( loop != $6 );
			      /* add onto DO list */
			      ADD2END(work_DOlst, work_Pobj);
			      /* sort out AABB */
			      do_box ( work_Pobj );
			  } else {
			      /*--- [CASE 2] : just the one iob_op */
			      popSTR( Sstack, $6->name );
			      popMLS( SPstack, $6->sprops );
			      /* add onto DO list */
			      ADD2END(work_DOlst, $6);
			  }
			  work_defining = FALSE;
			}
		  ;

obj_blk:	  COMPOBJ
		      Z1name
		      {
			work_Pobj = NEW(Tobject);
			DEFOPTRS(work_Pobj);
#ifdef __cplusplus
			work_Pobj->otype = object::comp;
#else
			work_Pobj->otype = comp;
#endif
			strcpy ( work_Pobj->name, $2 );
			pushMLS ( Pstack, work_Pobj );
		        unit_matrix ( work_TM1 );
		      }
		      options
		      {
			/* sort out trans matrix */
			memcpy ( work_Pobj->trans,
				 work_TM1, sizeof(Tmatrix) );

			/* sort out the CTM stack */
			if ( (MstackIndex) >= 0 ) {
			    exTMS ( Mstack, work_TM1 );
			    multiply_matrix ( work_TM1,
					      work_Pobj->trans,
					      work_TM1 );
			    pushTMS ( Mstack, work_TM1 );
			} else /* first matrix onto stack */
			    pushTMS ( Mstack, work_TM1 );

			/* sort out the surface properties */
			work_Pobj->sprops = work_sprops;
			pushMLS ( SPstack, work_sprops );
		      }
		      OMiob_ob
		      END
		      {
			ptr2obj	Ptop;

			/* sort out the CTM stack */
			popTMS ( Mstack, work_TM1 );

			/* sort out the parts/partof/sibling links */
			popMLS ( Pstack, work_Pobj );
			exPS ( Pstack, Ptop );
			if ( Ptop ) {
			    /* object is part of a COMPOBJ */
			    work_Pobj->partof = Ptop;
			    ADD2END( Ptop->parts, work_Pobj );
			} else {
			    /* bung on IOlst */
			    ADD2END( work_IOlst, work_Pobj );
			}

			/* sort out AABB for comp object */
			do_box ( work_Pobj );

			/* leaving COMPOBJ - reset surf props */
			popMLS ( SPstack, work_sprops );
			if ( SPstackIndex >= 0 )
			    exSPS ( SPstack, work_sprops );
			else
			    work_sprops.flags = 0;
		      }
		| obj_prims
		      {
			ptr2obj	Ptop;

			/* sort out the parts/partof/sibling links */
			exPS ( Pstack, Ptop );
			if ( Ptop ) {
			    /* object is part of a COMPOBJ */
			    $1->partof = Ptop;
			    ADD2END( Ptop->parts, $1 );
			} else {
			    /* bung on IOlst */
			    ADD2END( work_IOlst, $1 );
			}
		      }
		  ;

inst_obj_blk:	  INSTOBJ
			ID
			{
			  /*memcpy(work_Pobj, get_object($2), sizeof(Tobject));*/
			  work_Pobj=copy_object( get_object($2) );
			}
			Z1name
			{
			  strcpy ( work_Pobj->name, $4 );
			  unit_matrix ( work_TM1 );
			}
			options
			END
			{
			  /* surface properties */
			  PROPSmerge( work_Pobj->sprops,
				      work_sprops,
				      work_Pobj->sprops );
			  /* reset to what it was when we came in */
			  if ( SPstackIndex >= 0 )
			      exSPS( SPstack, work_sprops );
			  else
			      work_sprops.flags = 0;

			  /* sort out inst object's trans matrix */
			  multiply_matrix ( work_TM1,
					    work_Pobj->trans,
					    work_Pobj->trans );

			  /* sort out the object's bounding box */
			  EVAL_BOX ( work_Pobj );

			  /* sort out the parts/partof/sibling links */
			  if ( !work_defining ) {
			      ptr2obj	Ptop;

			      exPS ( Pstack, Ptop );
			      if ( Ptop ) {
				  /* object is part of a COMPOBJ */
				  work_Pobj->partof = Ptop;
				  ADD2END( Ptop->parts, work_Pobj );
			      } else
				  /* bung on IOlst */
				  ADD2END( work_IOlst, work_Pobj );
			  } else
			      $$ = work_Pobj;

			}
		  ;

debug:	 	  WALK
			{ walk_Olist ( work_IOlst, 0 ); }
		| WALK2
			{ walk_Olist ( work_DOlst, 0 ); }
		;


obj_prims:	  TRIANGLE
			Z1name
			{ pushSTR ( Sstack, $2 ); }
			OMpoint3	/* the vertices */
			Z1vnorm
			Z1fnorm
			{ unit_matrix ( work_TM1 ); }
			options
			END
			{ if ( $4 != 3 )	/* only "3" is OK */
			      yyerror ( "Triangle: Invalid No of verts" );
			  if ( $5 && $5 != 3 )	/* only 0 or 3 is OK */
			      yyerror ( "Triangle: Invalid No of v_norms" );
			  if ( $6 && $6 != 1 )	/* only 0 or 1 is OK */
			      yyerror ( "Triangle: Invalid No of f_norms" );
			  $$ = make_tri_obj( $5, $6 );

			  /*--- surface properties ---*/
			  $$->sprops = work_sprops;
			  /* reset to what it was when we came in */
			  if ( SPstackIndex >= 0 )
			      exSPS( SPstack, work_sprops );
			  else
			      work_sprops.flags = 0;

			  /*--- evaluate the object's bounding box ---*/
			  EVAL_BOX ( $$ );
			}	 
		| POLYGON
			Z1name
			{ 
			  pushSTR ( Sstack, $2 );
			  clean_v3lst();
			}
  			NUMBER			/* no of verts */
			OMpoint3		/* the verts */
			Z1vnorm
			Z1fnorm
			{ unit_matrix ( work_TM1 ); }
			options
			END
			{ if ( $4 != $5 )
			      yyerror ( "Polygon: Missing/extra vertices" );
			  if ( $6 && $4 != $6 )
			      yyerror ( "Polyg/vnorm: Missing/extra normals" );
			  if ( $7 && $4 != $7 )
			      yyerror ( "Polyg/fnorm: Missing/extra normals" );
			  $$ = make_polg_obj ( (int) $4, $6, $7 );

			  /*--- surface properties ---*/
			  $$->sprops = work_sprops;
			  /* reset to what it was when we came in */
			  if ( SPstackIndex >= 0 )
			      exSPS( SPstack, work_sprops );
			  else
			      work_sprops.flags = 0;

			  /*--- evaluate the object's bounding box ---*/
			  EVAL_BOX ( $$ );
			}
		| SPHERE
			Z1name
			{ pushSTR ( Sstack, $2 ); }
			point3			/* $4 - centre */
			NUMBER			/* $5 - radius */
			Z1up			/* $6 - up vector */
			Z1tpz			/* $7 - ThetaPhiZero vector */
			Z1theta			/* $8 - beg/end for theta */
			Z1phi			/* $9 - beg/end for phi */
			{ unit_matrix ( work_TM1 ); }
			options
			END
			{ if ( $5 == 0.0 )
			      yyerror ( "Sphere: zero radius" );
			  $$ = make_sph_obj ( &($4), $5, &($6),
 					      &($7), &($8), &($9) );

			  /*--- surface properties ---*/
			  $$->sprops = work_sprops;
			  /* reset to what it was when we came in */
			  if ( SPstackIndex >= 0 )
			      exSPS( SPstack, work_sprops );
			  else
			      work_sprops.flags = 0;

			  /*--- evaluate the object's bounding box ---*/
			  EVAL_BOX ( $$ );
			}
		| CYLINDER
			Z1name
			{
			  pushSTR ( Sstack, $2 );
			  clean_v3lst();
			}
			point3
			Z1capped
			point3
			Z1capped
			NUMBER		/* radius - $8 */
			Z1tz
			Z1theta
			{ unit_matrix ( work_TM1 ); }
			options
			END
			{ if ( $8 == 0.0 )
			      yyerror ( "Cylinder: Zero radius" );
			  $$ = make_cyl_obj ( &($4), $5, &($6),
						$7, $8, &($10) );

			  /*--- surface properties ---*/
			  $$->sprops = work_sprops;
			  /* reset to what it was when we came in */
			  if ( SPstackIndex >= 0 )
			      exSPS( SPstack, work_sprops );
			  else
			      work_sprops.flags = 0;

			  /*--- evaluate the object's bounding box ---*/
			  EVAL_BOX ( $$ );
			}
		| CONE
			Z1name
			{
			  pushSTR ( Sstack, $2 );
			  clean_v3lst();
			}
			point3
			Z1capped
			NUMBER
			point3
			Z1capped
			NUMBER
			Z1tz
			Z1theta
			{ unit_matrix ( work_TM1 ); }
			options
			END
			{ if ( !($6 || $9) )
			      yyerror ( "Cone: Zero radii" );
			  if ( (SGN($6) != SGN($9)) && $6 && $9 )
			      yyerror ( "Cone: Radii opposite signs" );
			  $$ = make_cone_obj ( &($4), $5, $6,
					       &($7), $8, $9, &($11) );

			  /*--- surface properties ---*/
			  $$->sprops = work_sprops;
			  /* reset to what it was when we came in */
			  if ( SPstackIndex >= 0 )
			      exSPS( SPstack, work_sprops );
			  else
			      work_sprops.flags = 0;

			  /*--- evaluate the object's bounding box ---*/
			  EVAL_BOX ( $$ );
			}
		| DISC
			Z1name
			{
			  pushSTR ( Sstack, $2 );
			  clean_v3lst();
			}
			point3			/* centre */
			NUMBER			/* radius */
			point3			/* surface normal */
			Z1tz
			Z1theta
			{ unit_matrix ( work_TM1 ); }
			options
			END
			{ if ( $5 <= 0.0 )
			      yyerror ( "Disc: -ve/zero radius" );
			  $$ = make_disc_obj ( &($4), $5, &($6), &($8) );

			  /*--- surface properties ---*/
			  $$->sprops = work_sprops;
			  /* reset to what it was when we came in */
			  if ( SPstackIndex >= 0 )
			      exSPS( SPstack, work_sprops );
			  else
			      work_sprops.flags = 0;

			  /*--- evaluate the object's bounding box ---*/
			  EVAL_BOX ( $$ );
			}
		| BOX
			Z1name
			{ pushSTR ( Sstack, $2 ); }
			point3
			point3
			Z1in
			{ unit_matrix ( work_TM1 ); }
			options
			END
			{
			  $$ = make_box_obj ( &($4), &($5), $6 );

			  /*--- surface properties ---*/
			  $$->sprops = work_sprops;
			  /* reset to what it was when we came in */
			  if ( SPstackIndex >= 0 )
			      exSPS( SPstack, work_sprops );
			  else
			      work_sprops.flags = 0;

			  /*--- evaluate the object's bounding box ---*/
			  EVAL_BOX ( $$ );
			}
		| NURBS
			Z1name
			{
			  pushSTR ( Sstack, $2 );
			  clean_Rlst();
			  clean_v3lst();
			  clean_v4lst();
			}
			ORDER
			NUMBER NUMBER			/* $5, $6 */
			CPTS
			NUMBER NUMBER			/* $8, $9 */
			KNOTS
			OPENBRACE OMreals CLOSEBRACE	/* U knot - $12 */
			OPENBRACE OMreals CLOSEBRACE	/* V knot - $15 */
			OMpoint4			/* ctrl pts - $17 */
			{ unit_matrix ( work_TM1 ); }
			options
			END
			{ if ( (int)($5 + $8) != $12 )
			      yyerror ( "NURBS: Missing/extra U-knots" ); 
			  if ( (int)($6 + $9) != $15 )
			      yyerror ( "NURBS: Missing/extra V-knots" );
			  if ( (int)($8 * $9) != $17 )
			      yyerror ( "NURBS: Missing/extra Ctrl pts" );
			  $$ = make_nurb_obj ( (int) $5, (int) $6, (int) $8, (int) $9 ); 

			  /*--- surface properties ---*/
			  $$->sprops = work_sprops;
			  /* reset to what it was when we came in */
			  if ( SPstackIndex >= 0 )
			      exSPS( SPstack, work_sprops );
			  else
			      work_sprops.flags = 0;

			  /*--- evaluate the object's bounding box ---*/
			  EVAL_BOX ( $$ );
			}
		| POLYHEDRON
			Z1name
			{
			  pushSTR ( Sstack, $2 );
			  clean_v3lst();
			}
			OPENBRACE
			OMpoint3		/* $5 - the vertices */
			CLOSEBRACE
			Z1vnorm			/* $7 - No of Vnorms */
			OPENBRACE
			{ clean_flst(); }
			OMvlists		/* faces; vertex index lists */
			CLOSEBRACE
			Z1fnorm			/* $11 - No of Fnorms */
			{ unit_matrix ( work_TM1 ); }
			options
			END
			{ if ( $7 && $5 != $7 )
			      yyerror ( "Polyh/vnorm: Missing/extra normals" );
			  if ( $12 && $10 != $12 )
			      yyerror ( "Polyh/fnorm: Missing/extra normals" );
			  $$ = make_polh_obj ( $5, $7, $10, $12 );

			  /*--- surface properties ---*/
			  $$->sprops = work_sprops;
			  /* reset to what it was when we came in */
			  if ( SPstackIndex >= 0 )
			      exSPS( SPstack, work_sprops );
			  else
			      work_sprops.flags = 0;

			  /*--- evaluate the object's bounding box ---*/
			  EVAL_BOX ( $$ );
			}
		  ;
			

/*===========================  L E V E L    5  ==============================*/

inst_props_blk:	  INSTPROPS
			ID
			{ get_props ( &work_sprops, $2 ); }
			;

/* A note on bit masks ... the "|=" operator sets the appropriate bit */
props_prims:	  EMISSION
			point3
		  { if ( $2.x < 0.0 || $2.y < 0.0 || $2.z < 0.0 )
			yyerror ( "Negative EMISSION" );
		    else {
			work_sprops.flags |= EMmask;
			work_sprops.emission = *(Tspectra *)(&($2));
		    }
		  }
		| ARC
			point3
		  { if ( OOR($2.x,0.0,1.0)	/* OOR; macro; Out Of Range */
		      || OOR($2.y,0.0,1.0)
		      || OOR($2.z,0.0,1.0) )
			yyerror ( "ARC out of range" );
		    else {
			work_sprops.flags |= ARCmask;
			work_sprops.arc = *(Tspectra *)(&($2));
		    }
		  }
		| DRC
			point3
		  { if ( OOR($2.x,0.0,1.0)
		      || OOR($2.y,0.0,1.0)
		      || OOR($2.z,0.0,1.0) )
			yyerror ( "DRC out of range" );
		    else {
			work_sprops.flags |= DRCmask;
			work_sprops.drc = *(Tspectra *)(&($2));
		    }
		  }
		| SRC
			point3
		  { if ( OOR($2.x,0.0,1.0)
		      || OOR($2.y,0.0,1.0)
		      || OOR($2.z,0.0,1.0) )
			yyerror ( "SRC out of range" );
		    else {
			work_sprops.flags |= SRCmask;
			work_sprops.src = *(Tspectra *)(&($2));
		    }
		  }
		| TRC
			point3
		  { if ( OOR($2.x,0.0,1.0)
		      || OOR($2.y,0.0,1.0)
		      || OOR($2.z,0.0,1.0) )
			yyerror ( "TRC out of range" );
		    else {
			work_sprops.flags |= TRCmask;
			work_sprops.trc = *(Tspectra *)(&($2));
		    }
		  }
		| PHEXP
			NUMBER
		  { if ( $2 < 0.0 )
			yyerror ( "Negative Phong exponent" );
		    else {
			work_sprops.flags |= PHmask;
			work_sprops.ph_exp = (int)$2;
		    }
		  }
		| INDEX
			point3
		{ if ( $2.x < 0.0 || $2.y < 0.0 || $2.z < 0.0 )
		      yyerror ( "Negative Refractive Index" );
		  else {
		      work_sprops.flags |= INmask;
		      work_sprops.index = *(Tspectra *)(&($2));
		  }
		}
		| TEXTURE
			Z1file
			ID
		{
		  if ( $2 )
		      work_sprops.flags |= FImask;
		  work_sprops.flags |= TEXmask;
		  strcpy ( work_sprops.texture, $3 );
		}
		  ;

inst_trans_blk:	INSTTRANS
			ID
			{
			  get_trans ( work_TM2, $2 );
			  multiply_matrix ( work_TM2,
					    work_TM1,
					    work_TM1 );
			}
		;

trans_prims:	SHIFT
			point3
			{
			  make_sh_mat ( &($2), work_TM2 );
			  multiply_matrix ( work_TM2,
					    work_TM1,
					    work_TM1 );
			}
		| SCALE
			point3
			{
			  make_sc_mat ( &($2), work_TM2 );
			  multiply_matrix ( work_TM2,
					    work_TM1,
					    work_TM1 );
			}
		| ROTATE
			ID
			{ work_char = $2[0]; }
			NUMBER
			{
			  make_rot_mat ( work_char, $4, work_TM2 );
			  multiply_matrix ( work_TM2,
					    work_TM1,
					    work_TM1 );
			}
		| MATRIX
			point4
			point4
			point4
			point4
			{
			  work_TM2[0][0] = $2.x;
		  	  work_TM2[0][1] = $2.y;
			  work_TM2[0][2] = $2.z;
			  work_TM2[0][3] = $2.w;

			  work_TM2[1][0] = $3.x;
			  work_TM2[1][1] = $3.y;
			  work_TM2[1][2] = $3.z;
			  work_TM2[1][3] = $3.w;

			  work_TM2[2][0] = $4.x;
			  work_TM2[2][1] = $4.y;
			  work_TM2[2][2] = $4.z;
			  work_TM2[2][3] = $4.w;

			  work_TM2[3][0] = $5.x;
			  work_TM2[3][1] = $5.y;
			  work_TM2[3][2] = $5.z;
			  work_TM2[3][3] = $5.w;

			  multiply_matrix ( work_TM2,
					    work_TM1,
					    work_TM1 );
			}
			;

/* 1 or more occ' of inst_obj_block | obj_prims .... these ARE self-linking */
OMiob_op:	  iob_op
		  {
		    $1->psib = $1->nsib = $1;
		    $$ = $1;
		  }
		| OMiob_op iob_op
		  {
		    ADD2END( $1, $2 );
		    $$ = $1;
		  }
		  ;


/* 0 or 1 occurences of 'name ID' */
Z1name:	  /* nothing */
			{ /* produces default names
				"defname1"..."defname9999" */
			  def_name ( $$ );
#ifdef VERBOSE
			  printf ( "DefName: %s\n", $$ );
#endif
			}
		| NAME ID
			{ strcpy ( $$, $2 ); }
			;

/* One or More inst_obj_blk | obj_blk .... NOT self-linking */
OMiob_ob:
		  iob_ob
		| OMiob_ob iob_ob
		  ;

iob_ob:		  inst_obj_blk
		| obj_blk
		  ;
			
/* an inst_obj_blk, or an object primitive */
iob_op:		  inst_obj_blk
		| obj_prims
		  ;


/* 0 or 1 vertex normal specs - return no of normals given */
Z1vnorm:	  /* nothing */
		  { $$ = 0; }
		| VNORM
			OMpoint3
		  { if ( $2 < 3 )
			yyerror ( "Missing points in V_NORM" );
		    $$ = $2;
		  }
		  ;

/* 0 or 1 facet normal specs - return no of normals given */
Z1fnorm:	  /* nothing */
		  { $$ = 0; }
		| FNORM
			OMpoint3
		  { if ( $2 < 1 )
			yyerror ( "Missing points in F_NORM" );
		    $$ = $2;
		  }
		  ;

/* One or More point3s - returns how many */
OMpoint3:	  point3
		{ $$ = 1;
		  add_point3 ( &($1) );
		}
		| OMpoint3 point3
		{ $$ = $1 + 1;
		  add_point3 ( &($2) );
		}
		  ;

Z1up:		  /* nothing - default value (0,0,1) */
		{ $$.x = 0.0;  $$.y = 0.0;  $$.z = 1.0; }
		| UP point3
		{ $$ = $2; }
		  ;

Z1tpz:		  /* nothing - default value (1,0,0) */
		{ $$.x = 1.0;  $$.y = 0.0;  $$.z = 0.0; }
		| TPZ point3
		{ $$ = $2; }
		  ;

Z1theta:	  /* nothing - default values (0,1) */
		{ $$.beg = 0.0;  $$.end = 1.0; }
		| THETA
			NUMBER
			NUMBER
		{ if ( OOR($2,0.0,1.0) || OOR($3,0.0,1.0) )
		      yyerror ( "Theta: out of range" );
		  $$.beg = $2;
		  $$.end = $3;
		}
		  ;

Z1phi:		  /* nothing - default values (0,1) */
		{ $$.beg = 0.0;  $$.end = 1.0; }
		| PHI
			NUMBER
			NUMBER
		{ if ( OOR($2,0.0,1.0) || OOR($3,0.0,1.0) )
		      yyerror ( "Phi: out of range" );
		  $$.beg = $2;
		  $$.end = $3;
		}
		  ;

Z1capped:	  /* nothing */
		{ $$ = FALSE; }
		| CAPPED
		{ $$ = TRUE; }
		  ;

Z1tz:		  /* nothing */
		| TZ point3
		{ add_point3 ( &($2) ); }
		  ;

Z1in:		  /* nothing */
		{ $$ = FALSE; }
		| IN
		{ $$ = TRUE; }
		  ;

options:	  /* nothing */
		| options single_option
		  ;

OMreals:	  NUMBER
		  { $$ = 1;
		    add_real ( $1 ); }
		| OMreals NUMBER
		  { $$ = $1 + 1;
		    add_real ( $2 );
		  }
		  ;

/* One or More point4s - returns how many */
OMpoint4:	  point4
		{ $$ = 1;
		  add_point4 ( &($1) );
		}
		| OMpoint4 point4
		{ $$ = $1 + 1;
		  add_point4 ( &($2) );
		}
		  ;

/* One or more vertex lists - to index into a list of vertices */
OMvlists:	  vlist
		  { $$ = 1; }
		| OMvlists vlist
		  { $$ = $1 + 1; }
		  ;

/*===========================  L E V E L    6  ==============================*/

/* 0 or 1 occurences of 'file' */
Z1file:	  	  /* nothing */
		  { $$ = FALSE; }
		| TEXFILE
		  { $$ = TRUE; }
		  ;

single_option:	  props_blk
		| inst_props_blk
		| trans_blk
		| inst_trans_blk
		| trans_prims
		| patches
		  ;

point4:		  point3
		  NUMBER
		  { $$.x = $1.x;
		    $$.y = $1.y;
		    $$.z = $1.z;
		    $$.w = $2;
		  }
		  ;

/* A vertex list */
vlist:		  OPENSQ
		  OMints
		  CLOSESQ
		{ /* add work_face info onto work_flst */
		  add2flst();
		}
		  ;

/*===========================  L E V E L    7  ==============================*/

props_blk:	  PROPS
			Z1name
			{ pushSTR ( Sstack, $2 ); }
			ZMipb
			ZMpp
			END
			{ popSTR ( Sstack, work_string );
			  if ( strncmp("default", work_string, 7) )
			      add_props( &work_sprops, $2 );
		       /* else
			      it was given a default name */
			}
			;

trans_blk:	  TRANS
			Z1name
			{
			  pushSTR ( Sstack, $2 );
			}
			ZMitb
			ZMtp
			END
			{
			  popSTR ( Sstack, work_string );
			  if ( strncmp ( "default", work_string, 7 ) )
			      add_trans ( work_TM1, work_string );
		       /* else
			      it was given a default name */
			}
		;

patches:	  PATCHES
			NUMBER
			{
			  work_patches.u = (int)$2;
		  	  work_patches.v = 0;
		  	  work_patches.w = 0;
			}
		| PATCHES
			NUMBER
			NUMBER
			{
			  work_patches.u = (int)$2;
			  work_patches.v = (int)$3;
			  work_patches.w = 0;
			}
		| PATCHES
			NUMBER
			NUMBER
			NUMBER
			{
			  work_patches.u = (int)$2;
			  work_patches.v = (int)$3;
			  work_patches.w = (int)$4;
			}
		  ;

OMints:		  NUMBER
		  { if ( $1 != (int)$1 )
		        yyerror ( "Must index vertex list with INTEGERS" );
		    add2face ( (int) $1, TRUE );
		  }
		| OMints NUMBER
		  { if ( $2 != (int)$2 )
		        yyerror ( "Index vertex list with INTEGERS" );
		    add2face ( (int) $2, FALSE );
		  }
		;
%%



/*===========================================================================*
 ************************ E N D    O F    Y A C C ****************************
 =============================================================================
 ************************* S T A R T    O F    C *****************************
 *===========================================================================*/

/*=== msdl() =================================================================*
  void msdl( FILE * )  parses the sdl file & then calls create
  object which calls the user written functions create_.....()
  It sets yyin, which is the FILE which Lex reads its input from
 *---------------------------------------------------------------------------*/

extern FILE *yyin;

void msdl(FILE *scene_file)
{
  /* yyin is the default file which Lex gets its input from */

  yyin = scene_file;
  if (!yyin) 
  { 
    fprintf ( stderr, "msdl(): Cannot read from that file\n");
    exit ( -1 );
  }

  /* Actually call the routines generated by yacc */

  yyparse();
  
  /* if there's a view specified create it */
  if( work_view!=NULL  )	
    create_view( work_view );
  /* if there's some instantiated object then create them!! */
  if( work_IOlst!=NULL )	
    create_object( work_IOlst );
  /* if there's some lights then create them */
  if( work_llst !=NULL )	
    create_lights( work_llst  );
}

/*===========================================================================*
  create_lights( Tlights* lite ) either recurses or calls low level application
  function.
 *---------------------------------------------------------------------------*/
void create_lights( Tlights* lite )
{
  do{
    switch( lite->ltype ){
	case pt:	create_point_lt( 	lite );	break;
	case dir:	create_directional_lt( 	lite );	break;
	case sp:	create_spot_lt(		lite );	break;
  }
  lite=lite->next;
  }while( lite!=NULL );
}

/*===========================================================================*
  create_object( Tobject* obj ) either recurses or calls low level application
  function.
 *---------------------------------------------------------------------------*/
void create_object( Tobject* objlist )
{
  Tobject* 	obj=objlist;
  Tobject*	next=objlist;
  Tmatrix	comptrans;

  do {
    memcpy( &comptrans, &obj->trans, sizeof( Tmatrix ) );
    acc_par_trans( obj, obj->trans, comptrans );

#ifndef __cplusplus
    switch( obj->otype ){
      case tri:		create_triangle(   obj, comptrans );   	break;
      case polg:	create_polygon(    obj, comptrans );	break;
      case sph:		create_sphere( 	   obj, comptrans );	break;
      case cyl:		create_cylinder(   obj, comptrans );	break;
      case cone:	create_cone( 	   obj, comptrans );	break;
      case disc:	create_disc(       obj, comptrans );	break;
      case ebox:	create_box(        obj, comptrans );	break;
      case nurb:	create_nurb(       obj, comptrans );	break;
      case polh:	create_polyhedron( obj, comptrans );	break;
      case comp:	create_comp(	   obj, comptrans );
			create_object(     obj->parts );
			end_comp();				break;
    }
#else
    // Do C++ variant
    switch( obj->otype ){
      case (object::tri):	create_triangle(obj, comptrans );   	break;
      case (object::polg):	create_polygon(obj, comptrans );	break;
      case (object::sph):	create_sphere(obj, comptrans );	break;
      case (object::cyl):	create_cylinder(obj, comptrans );	break;
      case (object::cone):	create_cone(obj, comptrans );	break;
      case (object::disc):	create_disc(obj, comptrans );	break;
      case (object::ebox):	create_box(obj, comptrans );	break;
      case (object::nurb):	create_nurb(obj, comptrans );	break;
      case (object::polh):	create_polyhedron(obj, comptrans );	break;
      case (object::comp):	create_comp(obj, comptrans );
	create_object(obj->parts );
	end_comp();				break;
    }
#endif
    obj=obj->nsib;
  } while( obj!=objlist );
}

/*===========================================================================*
 err_malloc: calls malloc() with the given size and returns the pointer
 returned by malloc() if all is well, and prints an error message otherwise.
 IN:	size_t		size		How much space do I want ?
 *---------------------------------------------------------------------------*/
void *err_malloc ( unsigned size )
{
#ifdef __cplusplus
    void *result = new char[size];
#else
    void *result = (char *) malloc(size);
#endif
    if ( result == NULL ) {
	fprintf ( stderr, "Malloc failed\n" );
	exit ( -1 );
    } else
	return ( result );     
}
/*--- end of err_malloc -----------------------------------------------------*/


/*===========================================================================*
 yyerror: What YACC code would be complete without it ! Takes an error message
 and prints it, together with the line number where I think it took place.
 *---------------------------------------------------------------------------*/
void yyerror ( char *errmess )
{
    extern int lineno;

    create_problem(errmess, lineno);
}
/*--- end of yyerror --------------------------------------------------------*/


/*===========================================================================*
 def_name: given a char* to fill, it generates a default name and fills the
 string

 I/O:	char *		defname		the name I'm filling
 *---------------------------------------------------------------------------*/
void def_name ( char *defname )
{
    static int count = 0;

    count++;
    if ( count > 9999 )
	yyerror ( "def_name: Too many default names" );
    sprintf ( defname, "default%.4i", count );
}
/*--- end of def_name -------------------------------------------------------*/


void cleanup()
{
    /* free everything in sight */
  clean_v4lst();
  clean_v3lst();
  clean_flst();
  clean_Rlst();
}


void remove_light_list( Tlights* lite ) /* red hot routine that's never called */
{
  Tlights	*trash,
		*nxt;
  for( trash = work_llst ; trash != NULL ; trash=nxt ){
	nxt = trash->next;
	free( trash );
  }
  work_llst = NULL;
}

void remove_obj_lst( Tobject* lst )	/* red hot routine that's never called */
{
  Tobject*	next=lst;

  do{
    remove_obj( next );
    next=next->nsib;
  }while( next!=lst );
}

void remove_obj( Tobject* obj )
{
  Tobject*	part=NULL;
  Tpint		i;

#ifndef __cplusplus
  if( obj->otype==comp ){
    /* it's a composite */
    part = obj->parts;
    do {	/* remove each part */
      remove_obj( part );
      /* next part ... */
      part = part->nsib;
    } while ( part != obj->parts );		
    free( obj );
  } else {
    /* It's a primative it's easy to remove */
    /* remove object specific stuff */
    switch( obj->otype ){
	case polg:
		free(obj->spec.polg.vvnfn.data);
		break;
	case tri:
		free(obj->spec.tri.vvnfn.data);
		break;
	case polh:
		free(obj->spec.polh.vvnfn.data);
		/* remove the face details */
		for( i=0 ; i<obj->spec.polh.flist.no_faces ; i++ ){
		  /* make enough room for that many verts in the index array */
		  free(obj->spec.polh.flist.face[i]->vert_index);
		  free(obj->spec.polh.flist.face[i]);
		  /* repeat for the next face */
		}
		free(obj->spec.polh.flist.face);
		break;
	case nurb:
		free(obj->spec.nurb.knots_u);
		free(obj->spec.nurb.knots_v);
		free(obj->spec.nurb.cpts);
		break;
      }
    free( obj );
  }
#else
  if( obj->otype==object::comp ){
    /* it's a composite */
    part = obj->parts;
    do {	/* remove each part */
      remove_obj( part );
      /* next part ... */
      part = part->nsib;
    } while ( part != obj->parts );		
    free( obj );
  } else {
    /* It's a primative it's easy to remove */
    /* remove object specific stuff */
    switch( obj->otype ){
	case object::polg:
		free(obj->spec.polg.vvnfn.data);
		break;
	case object::tri:
		free(obj->spec.tri.vvnfn.data);
		break;
	case object::polh:
		free(obj->spec.polh.vvnfn.data);
		/* remove the face details */
		for( i=0 ; i<obj->spec.polh.flist.no_faces ; i++ ){
		  /* make enough room for that many verts in the index array */
		  free(obj->spec.polh.flist.face[i]->vert_index);
		  free(obj->spec.polh.flist.face[i]);
		  /* repeat for the next face */
		}
		free(obj->spec.polh.flist.face);
		break;
	case object::nurb:
		free(obj->spec.nurb.knots_u);
		free(obj->spec.nurb.knots_v);
		free(obj->spec.nurb.cpts);
		break;
      }
    free( obj );
  }
#endif
}
